@@ -1,5 +1,5 @@
 #!/usr/bin/env python
-# Copyright 2010 Google Inc.
+# Copyright 2010 Google Inc. All Rights Reserved.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -43,11 +43,11 @@ from mapreduce import base_handler
 from mapreduce import context
 from mapreduce import errors
 from mapreduce import input_readers
-from mapreduce import map_job
 from mapreduce import model
 from mapreduce import operation
 from mapreduce import parameters
 from mapreduce import util
+from mapreduce.api import map_job
 from google.appengine.runtime import apiproxy_errors
 
 # pylint: disable=g-import-not-at-top
@@ -102,19 +102,26 @@ def _run_task_hook(hooks, method, task, queue_name):
 class MapperWorkerCallbackHandler(base_handler.HugeTaskHandler):
   """Callback handler for mapreduce worker task."""
 
+  # These directives instruct self.__return() how to set state and enqueue task.
   _TASK_DIRECTIVE = util._enum(
       # Task is running as expected.
       PROCEED_TASK="proceed_task",
-      # Retry task. Don't change payload or datastore.
+      # Need to retry task. Lock was NOT acquired when the error occur.
+      # Don't change payload or datastore.
       RETRY_TASK="retry_task",
-      # Drop the task (due to duplicated task). Log permanent drop.
+      # Need to retry task. Lock was acquired when the error occurr.
+      # Don't change payload or datastore.
+      RETRY_SLICE="retry_slice",
+      # Drop the task (due to duplicated task). Must log permanent drop.
       DROP_TASK="drop_task",
       # See handlers.MapperWorkerCallbackHandler._attempt_slice_recovery.
       RECOVER_SLICE="recover_slice",
-      # Shard should be retried.
+      # Need to retry the shard.
       RETRY_SHARD="retry_shard",
-      # Drop task and fail the shard. Log permanent failure.
-      FAIL_TASK="fail_task")
+      # Need to drop task and fail the shard. Log permanent failure.
+      FAIL_TASK="fail_task",
+      # Need to abort the shard.
+      ABORT_SHARD="abort_shard")
 
   def __init__(self, *args):
     """Constructor."""
@@ -206,8 +213,7 @@ class MapperWorkerCallbackHandler(base_handler.HugeTaskHandler):
     # See model.ShardState doc.
     if shard_state.slice_start_time:
       countdown = self._wait_time(shard_state,
-                                  parameters.config._LEASE_GRACE_PERIOD +
-                                  parameters.config._SLICE_DURATION_SEC)
+                                  parameters._LEASE_DURATION_SEC)
       if countdown > 0:
         logging.warning(
             "Last retry of slice %s-%s may be still running."
@@ -221,7 +227,7 @@ class MapperWorkerCallbackHandler(base_handler.HugeTaskHandler):
       # lease could have expired. Verify with logs API.
       else:
         if self._wait_time(shard_state,
-                           parameters.config._REQUEST_EVENTUAL_TIMEOUT):
+                           parameters._MAX_LEASE_DURATION_SEC):
           if not self._has_old_request_ended(shard_state):
             logging.warning(
                 "Last retry of slice %s-%s is still in flight with request_id "
@@ -233,7 +239,7 @@ class MapperWorkerCallbackHandler(base_handler.HugeTaskHandler):
               "Last retry of slice %s-%s has no log entry and has"
               "timed out after %s seconds",
               tstate.shard_id, tstate.slice_id,
-              parameters.config._REQUEST_EVENTUAL_TIMEOUT)
+              parameters._MAX_LEASE_DURATION_SEC)
 
     # Lease expired or slice_start_time not set.
     config = util.create_datastore_write_config(tstate.mapreduce_spec)
@@ -244,6 +250,9 @@ class MapperWorkerCallbackHandler(base_handler.HugeTaskHandler):
       If failed for any reason, raise error to retry the task (hence all
       the previous validation code). The task would die naturally eventually.
 
+      Raises:
+        Rollback: If the shard state is missing.
+
       Returns:
         A _TASK_DIRECTIVE enum.
       """
@@ -387,47 +396,57 @@ class MapperWorkerCallbackHandler(base_handler.HugeTaskHandler):
     if shard_state:
       is_this_a_retry = shard_state.acquired_once
     task_directive = self._try_acquire_lease(shard_state, tstate)
-    if task_directive == self._TASK_DIRECTIVE.RETRY_TASK:
-      return self.retry_task()
-    if task_directive == self._TASK_DIRECTIVE.DROP_TASK:
-      return
+    if task_directive in (self._TASK_DIRECTIVE.RETRY_TASK,
+                          self._TASK_DIRECTIVE.DROP_TASK):
+      return self.__return(shard_state, tstate, task_directive)
     assert task_directive == self._TASK_DIRECTIVE.PROCEED_TASK
 
     # Abort shard if received signal.
     if control and control.command == model.MapreduceControl.ABORT:
-      logging.info("Abort command received by shard %d of job '%s'",
-                   shard_state.shard_number, shard_state.mapreduce_id)
-      # NOTE: When aborting, specifically do not finalize the output writer
-      # because it might be in a bad state.
-      shard_state.set_for_abort()
-      shard_state.put(config=util.create_datastore_write_config(spec))
-      return
+      task_directive = self._TASK_DIRECTIVE.ABORT_SHARD
+      return self.__return(shard_state, tstate, task_directive)
+
+    # Retry shard if user disabled slice retry.
+    if (is_this_a_retry and
+        parameters.config.TASK_MAX_DATA_PROCESSING_ATTEMPTS <= 1):
+      task_directive = self._TASK_DIRECTIVE.RETRY_SHARD
+      return self.__return(shard_state, tstate, task_directive)
 
     # TODO(user): Find a better way to set these per thread configs.
     # E.g. what if user change it?
     util._set_ndb_cache_policy()
-    if cloudstorage:
-      cloudstorage.set_default_retry_params(
-          cloudstorage.RetryParams(
-              urlfetch_timeout=parameters._GCS_URLFETCH_TIMEOUT_SEC))
 
+    job_config = map_job.JobConfig._to_map_job_config(
+        spec,
+        os.environ.get("HTTP_X_APPENGINE_QUEUENAME"))
+    job_context = map_job.JobContext(job_config)
+    self.shard_context = map_job.ShardContext(job_context, shard_state)
+    self.slice_context = map_job.SliceContext(self.shard_context,
+                                              shard_state,
+                                              tstate)
     try:
       if is_this_a_retry:
         task_directive = self._attempt_slice_recovery(shard_state, tstate)
+        if task_directive != self._TASK_DIRECTIVE.PROCEED_TASK:
+          return self.__return(shard_state, tstate, task_directive)
 
-      if task_directive == self._TASK_DIRECTIVE.PROCEED_TASK:
-        finished_shard = self._process_inputs(
-            tstate.input_reader, shard_state, tstate, ctx)
-        if finished_shard:
-          # Since there was no exception raised, we can finalize output writer
-          # safely. Otherwise writer might be stuck in some bad state.
-          if tstate.output_writer:
-            # It's possible that finalization is successful but
-            # saving state failed. In this case this shard will retry upon
-            # finalization error.
-            # TODO(user): make finalize method idempotent!
-            tstate.output_writer.finalize(ctx, shard_state)
-          shard_state.set_for_success()
+      if isinstance(tstate.handler, map_job.Mapper):
+        if tstate.slice_id == 0:
+          tstate.handler.begin_shard(self.shard_context)
+        tstate.handler.begin_slice(self.slice_context)
+
+      finished_shard = self._process_inputs(
+          tstate.input_reader, shard_state, tstate, ctx)
+      if finished_shard:
+        # Since there was no exception raised, we can finalize output writer
+        # safely. Otherwise writer might be stuck in some bad state.
+        if tstate.output_writer:
+          # It's possible that finalization is successful but
+          # saving state failed. In this case this shard will retry upon
+          # finalization error.
+          # TODO(user): make finalize method idempotent!
+          tstate.output_writer.finalize(ctx, shard_state)
+        shard_state.set_for_success()
     # pylint: disable=broad-except
     except Exception, e:
       logging.warning("Shard %s got error.", shard_state.shard_id)
@@ -438,8 +457,12 @@ class MapperWorkerCallbackHandler(base_handler.HugeTaskHandler):
         logging.error("Got FailJobError.")
         task_directive = self._TASK_DIRECTIVE.FAIL_TASK
       else:
-        task_directive = self._TASK_DIRECTIVE.RETRY_TASK
+        task_directive = self._TASK_DIRECTIVE.RETRY_SLICE
 
+    self.__return(shard_state, tstate, task_directive)
+
+  def __return(self, shard_state, tstate, task_directive):
+    """Handler should always call this as the last statement."""
     task_directive = self._set_state(shard_state, tstate, task_directive)
     self._save_state_and_schedule_next(shard_state, tstate, task_directive)
 
@@ -510,9 +533,14 @@ class MapperWorkerCallbackHandler(base_handler.HugeTaskHandler):
         break
 
     # Flush context and its pools.
-    operation.counters.Increment(
+    self.slice_context.incr(
         context.COUNTER_MAPPER_WALLTIME_MS,
-        int((self._time() - self._start_time)*1000))(ctx)
+        int((self._time() - self._start_time)*1000))
+
+    if isinstance(tstate.handler, map_job.Mapper):
+      tstate.handler.end_slice(self.slice_context)
+      if finished_shard:
+        tstate.handler.end_shard(self.shard_context)
     ctx.flush()
 
     return finished_shard
@@ -532,26 +560,29 @@ class MapperWorkerCallbackHandler(base_handler.HugeTaskHandler):
       True if scan should be continued, False if scan should be stopped.
     """
     if data is not input_readers.ALLOW_CHECKPOINT:
-      ctx.counters.increment(context.COUNTER_MAPPER_CALLS)
+      self.slice_context.incr(context.COUNTER_MAPPER_CALLS)
 
       handler = transient_shard_state.handler
 
-      if input_reader.expand_parameters:
-        result = handler(*data)
+      if isinstance(handler, map_job.Mapper):
+        handler(self.slice_context, data)
       else:
-        result = handler(data)
+        if input_reader.expand_parameters:
+          result = handler(*data)
+        else:
+          result = handler(data)
 
-      if util.is_generator(result):
-        for output in result:
-          if isinstance(output, operation.Operation):
-            output(ctx)
-          else:
-            output_writer = transient_shard_state.output_writer
-            if not output_writer:
-              logging.warning(
-                  "Handler yielded %s, but no output writer is set.", output)
+        if util.is_generator(result):
+          for output in result:
+            if isinstance(output, operation.Operation):
+              output(ctx)
             else:
-              output_writer.write(output)
+              output_writer = transient_shard_state.output_writer
+              if not output_writer:
+                logging.warning(
+                    "Handler yielded %s, but no output writer is set.", output)
+              else:
+                output_writer.write(output)
 
     if self._time() - self._start_time >= parameters.config._SLICE_DURATION_SEC:
       return False
@@ -566,11 +597,24 @@ class MapperWorkerCallbackHandler(base_handler.HugeTaskHandler):
       task_directive: self._TASK_DIRECTIVE for current shard.
 
     Returns:
-      A _TASK_DIRECTIVE enum. PROCEED_TASK if task should proceed normally.
-    RETRY_SHARD if shard should be retried. RETRY_TASK if slice should be
-    retried. FAIL_TASK if task should. RECOVER_SLICE if slice should be
-    recovered.
+      A _TASK_DIRECTIVE enum.
+      PROCEED_TASK if task should proceed normally.
+      RETRY_SHARD if shard should be retried.
+      RETRY_SLICE if slice should be retried.
+      FAIL_TASK if sahrd should fail.
+      RECOVER_SLICE if slice should be recovered.
+      ABORT_SHARD if shard should be aborted.
+      RETRY_TASK if task should be retried.
+      DROP_TASK if task should be dropped.
     """
+    if task_directive in (self._TASK_DIRECTIVE.RETRY_TASK,
+                          self._TASK_DIRECTIVE.DROP_TASK):
+      return task_directive
+
+    if task_directive == self._TASK_DIRECTIVE.ABORT_SHARD:
+      shard_state.set_for_abort()
+      return task_directive
+
     if task_directive == self._TASK_DIRECTIVE.PROCEED_TASK:
       shard_state.advance_for_next_slice()
       tstate.advance_for_next_slice()
@@ -581,12 +625,13 @@ class MapperWorkerCallbackHandler(base_handler.HugeTaskHandler):
       shard_state.advance_for_next_slice(recovery_slice=True)
       return task_directive
 
-    if task_directive == self._TASK_DIRECTIVE.RETRY_TASK:
+    if task_directive == self._TASK_DIRECTIVE.RETRY_SLICE:
       task_directive = self._attempt_slice_retry(shard_state, tstate)
     if task_directive == self._TASK_DIRECTIVE.RETRY_SHARD:
       task_directive = self._attempt_shard_retry(shard_state, tstate)
     if task_directive == self._TASK_DIRECTIVE.FAIL_TASK:
       shard_state.set_for_failure()
+
     return task_directive
 
   def _save_state_and_schedule_next(self, shard_state, tstate, task_directive):
@@ -601,12 +646,22 @@ class MapperWorkerCallbackHandler(base_handler.HugeTaskHandler):
       shard_state: model.ShardState for current shard.
       tstate: model.TransientShardState for current shard.
       task_directive: enum _TASK_DIRECTIVE.
+
+    Returns:
+      The task to retry if applicable.
     """
     spec = tstate.mapreduce_spec
 
-    if task_directive == self._TASK_DIRECTIVE.RETRY_TASK:
+    if task_directive == self._TASK_DIRECTIVE.DROP_TASK:
+      return
+    if task_directive in (self._TASK_DIRECTIVE.RETRY_SLICE,
+                          self._TASK_DIRECTIVE.RETRY_TASK):
       # Set HTTP code to 500.
       return self.retry_task()
+    elif task_directive == self._TASK_DIRECTIVE.ABORT_SHARD:
+      logging.info("Aborting shard %d of job '%s'",
+                   shard_state.shard_number, shard_state.mapreduce_id)
+      task = None
     elif task_directive == self._TASK_DIRECTIVE.FAIL_TASK:
       logging.critical("Shard %s failed permanently.", shard_state.shard_id)
       task = None
@@ -634,6 +689,7 @@ class MapperWorkerCallbackHandler(base_handler.HugeTaskHandler):
 
     @db.transactional(retries=5)
     def _tx():
+      """The Transaction helper."""
       fresh_shard_state = model.ShardState.get_by_shard_id(tstate.shard_id)
       if not fresh_shard_state:
         raise db.Rollback()
@@ -758,7 +814,7 @@ class MapperWorkerCallbackHandler(base_handler.HugeTaskHandler):
       tstate: model.TransientShardState for current shard.
 
     Returns:
-      A _TASK_DIRECTIVE enum. RETRY_TASK if slice should be retried.
+      A _TASK_DIRECTIVE enum. RETRY_SLICE if slice should be retried.
     RETRY_SHARD if shard retry should be attempted.
     """
     if (shard_state.slice_retries + 1 <
@@ -778,7 +834,7 @@ class MapperWorkerCallbackHandler(base_handler.HugeTaskHandler):
       # in log.
       sys.exc_clear()
       self._try_free_lease(shard_state, slice_retry=True)
-      return self._TASK_DIRECTIVE.RETRY_TASK
+      return self._TASK_DIRECTIVE.RETRY_SLICE
 
     if parameters.config.TASK_MAX_DATA_PROCESSING_ATTEMPTS > 0:
       logging.warning("Slice attempt %s exceeded %s max attempts.",
@@ -848,7 +904,7 @@ class MapperWorkerCallbackHandler(base_handler.HugeTaskHandler):
         tstate.slice_id,
         tstate.retries)
 
-    headers = util._get_task_headers(tstate.mapreduce_spec)
+    headers = util._get_task_headers(tstate.mapreduce_spec.mapreduce_id)
     headers[util._MR_SHARD_ID_TASK_HEADER] = tstate.shard_id
 
     worker_task = model.HugeTask(
@@ -1072,6 +1128,7 @@ class ControllerCallbackHandler(base_handler.HugeTaskHandler):
     else:
       @db.transactional(retries=5)
       def _put_state():
+        """The helper for storing the state."""
         fresh_state = model.MapreduceState.get_by_job_id(spec.mapreduce_id)
         # We don't check anything other than active because we are only
         # updating stats. It's OK if they are briefly inconsistent.
@@ -1126,12 +1183,13 @@ class ControllerCallbackHandler(base_handler.HugeTaskHandler):
     if done_callback:
       done_task = taskqueue.Task(
           url=done_callback,
-          headers=util._get_task_headers(mapreduce_spec,
+          headers=util._get_task_headers(mapreduce_spec.mapreduce_id,
                                          util.CALLBACK_MR_ID_TASK_HEADER),
           method=mapreduce_spec.params.get("done_callback_method", "POST"))
 
     @db.transactional(retries=5)
     def _put_state():
+      """Helper to store state."""
       fresh_state = model.MapreduceState.get_by_job_id(
           mapreduce_spec.mapreduce_id)
       if not fresh_state.active:
@@ -1162,7 +1220,8 @@ class ControllerCallbackHandler(base_handler.HugeTaskHandler):
     """Compute single controller task name.
 
     Args:
-      transient_shard_state: an instance of TransientShardState.
+      mapreduce_spec: specification of the mapreduce.
+      serial_id: id of the invocation as int.
 
     Returns:
       task name which should be used to process specified shard/slice.
@@ -1217,7 +1276,7 @@ class ControllerCallbackHandler(base_handler.HugeTaskHandler):
         name=task_name, params=task_params,
         countdown=parameters.config._CONTROLLER_PERIOD_SEC,
         parent=mapreduce_state,
-        headers=util._get_task_headers(mapreduce_spec))
+        headers=util._get_task_headers(mapreduce_spec.mapreduce_id))
 
     if not _run_task_hook(mapreduce_spec.get_hooks(),
                           "enqueue_controller_task",
@@ -1472,7 +1531,7 @@ class StartJobHandler(base_handler.PostJsonHandler):
         "params_validator", "params.")
 
     # Default values.
-    mr_params = map_job.MapJobConfig._get_default_mr_params()
+    mr_params = map_job.JobConfig._get_default_mr_params()
     mr_params.update(params)
     if "queue_name" in mapper_params:
       mr_params["queue_name"] = mapper_params["queue_name"]
@@ -1508,6 +1567,9 @@ class StartJobHandler(base_handler.PostJsonHandler):
     Raises:
       Any exception raised by the 'params_validator' request parameter if
       the params fail to validate.
+
+    Returns:
+      The user parameters.
     """
     params_validator = self.request.get(validator_parameter)
 
@@ -1639,11 +1701,12 @@ class StartJobHandler(base_handler.PostJsonHandler):
                         eta,
                         countdown,
                         queue_name):
+    """Enqueues a new kickoff task."""
     params = {"mapreduce_id": mapreduce_spec.mapreduce_id}
     # Task is not named so that it can be added within a transaction.
     kickoff_task = taskqueue.Task(
         url=base_path + "/kickoffjob_callback/" + mapreduce_spec.mapreduce_id,
-        headers=util._get_task_headers(mapreduce_spec),
+        headers=util._get_task_headers(mapreduce_spec.mapreduce_id),
         params=params,
         eta=eta,
         countdown=countdown)
@@ -1664,7 +1727,8 @@ class FinalizeJobHandler(base_handler.TaskQueueHandler):
     mapreduce_id = self.request.get("mapreduce_id")
     mapreduce_state = model.MapreduceState.get_by_job_id(mapreduce_id)
     if mapreduce_state:
-      config=util.create_datastore_write_config(mapreduce_state.mapreduce_spec)
+      config = (
+          util.create_datastore_write_config(mapreduce_state.mapreduce_spec))
       keys = [model.MapreduceControl.get_key_by_job_id(mapreduce_id)]
       for ss in model.ShardState.find_all_by_mapreduce_state(mapreduce_state):
         keys.extend(list(
@@ -1686,7 +1750,7 @@ class FinalizeJobHandler(base_handler.TaskQueueHandler):
         url=(mapreduce_spec.params["base_path"] + "/finalizejob_callback/" +
              mapreduce_spec.mapreduce_id),
         params={"mapreduce_id": mapreduce_spec.mapreduce_id},
-        headers=util._get_task_headers(mapreduce_spec))
+        headers=util._get_task_headers(mapreduce_spec.mapreduce_id))
     queue_name = util.get_queue_name(None)
     if not _run_task_hook(mapreduce_spec.get_hooks(),
                           "enqueue_controller_task",
