@@ -691,17 +691,14 @@ class ShufflePipeline(pipeline_base.PipelineBase):
   """
 
   def run(self, job_name, filenames, shards=None):
-    if files.shuffler.available():
-      yield _ShuffleServicePipeline(job_name, filenames)
-    else:
-      hashed_files = yield _HashPipeline(job_name, filenames, shards=shards)
-      sorted_files = yield _SortChunksPipeline(job_name, hashed_files)
-      temp_files = [hashed_files, sorted_files]
+    hashed_files = yield _HashPipeline(job_name, filenames, shards=shards)
+    sorted_files = yield _SortChunksPipeline(job_name, hashed_files)
+    temp_files = [hashed_files, sorted_files]
 
-      merged_files = yield _MergePipeline(job_name, sorted_files)
+    merged_files = yield _MergePipeline(job_name, sorted_files)
 
-      with pipeline.After(merged_files):
-        all_temp_files = yield pipeline_common.Extend(*temp_files)
-        yield mapper_pipeline._CleanupPipeline(all_temp_files)
+    with pipeline.After(merged_files):
+      all_temp_files = yield pipeline_common.Extend(*temp_files)
+      yield mapper_pipeline._CleanupPipeline(all_temp_files)
 
-      yield pipeline_common.Return(merged_files)
+    yield pipeline_common.Return(merged_files)
